//------------------------------------------------------------------------------
//
//  npnode.c
//  npe
//
//  Created by Shane Saxon on 2/19/10.
//  Copyright Saxon Digital 2010. All rights reserved.
//
//------------------------------------------------------------------------------

#include "npnode.h"

#include <stdlib.h>

#include "../npctrl.h"


void InitNodeSurface (NPnodePtr node);
void InitNodePoints (NPnodePtr node);
void InitNodePin (NPnodePtr node);
void InitNodeVideo (NPnodePtr node);
void InitNodeGrid (NPnodePtr node);


//------------------------------------------------------------------------------
void npInitNode (void* dataRef)
{
	int i = 0;

	pData data = (pData) dataRef;

	for (i=0; i < kNPnodeMax; i++)
		data->map.node[i] = NULL;
}


//------------------------------------------------------------------------------
void npCloseNode(void* dataRef)
{
	int i = 0;

	pData data = (pData) dataRef;

	for (i=0; i < data->map.nodeCount; i++)
		npNodeDelete (data->map.node[i], dataRef);
}


// should make this static and add locking for delete to be thread safe,      debug zz
//------------------------------------------------------------------------------
void* npNodeNew (int nodeType, void* dataRef)
{
	int nodeIndex = 0;
	NPnodePtr node = NULL;

	pData data = (pData) dataRef;

	//return if at the node max
	if ( data->map.nodeCount == kNPnodeMax )
		return NULL;							// add error nodemax , debug zz

	//init the node
	node = (NPnodePtr) malloc (sizeof(NPnode));
	if (data == NULL)
	{
		printf ("\n 4219 error malloc failed cannot write file \n");
		return NULL;
	}

	node->id = npNewNodeID();
	npInitNodeDefault (node);	

	node->type = nodeType;				//check to make sure inbounds, debug, zz

	switch (nodeType)
	{
		case kNodeCamera :	npInitNodeCamera (node); break;
		case kNodeGrid :	InitNodeGrid (node); break;
		case kNodeVideo :	InitNodeVideo (node); break;
		case kNodeSurface :	InitNodeSurface (node); break;
		case kNodePoints :	InitNodePoints (node); break;
		case kNodePin :		
			InitNodePin (node); 
			npNodeNewChild (node, dataRef);	// should not need this, w/o breaks bad, debug, zz
			node->translate.x += 5.0f * data->map.nodeCount -35.0f;
			break;
		case kNodePinTorus :
			InitNodePin (node);
			node->surfaceType = kSurfaceTorus;
			return (void*)node;
			break;

		default : 
			break;
	}

	data->map.node[data->map.nodeCount] = node;
	data->map.nodeCount++;
	data->map.nodeIndex = data->map.nodeCount - 1;


	data->map.activeNode = node;
	data->map.selectedNode = data->map.activeNode;
	data->map.selectedIndex = data->map.nodeIndex;

	return (void*)node;
}

					// add a primary torus child to the trunk,
					//					make this part of the npNodeNew.. debug, zz
//------------------------------------------------------------------------------
void* npNodeNewChild (NPnodePtr node, void* dataRef)
{
	NPnodePtr nodeChild = NULL;

	// we only allow creating children of primary toroid,				debug, zz
	if (node->branchLevel < 1 && node->childCount > 0)		// allow for multiple primaries
		node = node->child[node->childIndex];

	// create and add child to level 1
	if (node->childCount >= kNPnodeChildMax)
		return NULL;

	nodeChild = npNodeNew (kNodePinTorus, dataRef);
					
	nodeChild->parent = node;
	nodeChild->branchLevel = node->branchLevel + 1;

	// distribute the children at 25 deg spacing... is 0.44 radians
	nodeChild->rotateRad.x = 0.44f * (float)node->childCount;

	nodeChild->shaderType = kNPglutSolidTorus;

	// attach child to parent
	node->child[node->childCount] = nodeChild;
	node->childCount++;
	node->childIndex = node->childCount - 1;

	return (void*)nodeChild;
}

															// add travers tree delete...
//------------------------------------------------------------------------------
void npNodeDelete (NPnodePtr node, void* dataRef)
{
	int i = 0;

	pData data = (pData) dataRef;
	NPnodePtr parent = NULL;

	if (data->map.nodeIndex < 3 || node == NULL)	//do not delete the root grid or camera
		return;

	if (node->branchLevel > 0)
	{
		parent = node->parent;

		// don't delete the last torus...			// fix thi
		if (node->branchLevel == 1)
		{
			data->map.selectedNode = parent;
			data->map.activeNode = data->map.selectedNode;
			return;
		}

		// compact the gap, optimize with block copy later...		zz
		parent->childCount--;
		for (i = parent->childIndex; i < parent->childCount; i++)
			parent->child[i] = parent->child[i+1];

		parent->child[parent->childCount] = NULL;

		if (parent->childIndex >= parent->childCount
			&& parent->childCount > 0 )
			parent->childIndex = parent->childCount - 1;

		if (parent->childCount <= 0)
		{
			data->map.selectedNode = parent;
			data->map.activeNode = data->map.selectedNode;
		}
		else
		{
			data->map.selectedNode = parent->child[parent->childIndex];
			data->map.activeNode = data->map.selectedNode;
		}
	}
	else
	{
		// compact the gap, optimize with block copy later...		zz
		data->map.nodeCount--;
		for (i = data->map.nodeIndex; i < data->map.nodeCount; i++)
			data->map.node[i] = data->map.node[i+1];

		data->map.node[data->map.nodeCount] = NULL;

		if (data->map.nodeIndex >= data->map.nodeCount)
			data->map.nodeIndex = data->map.nodeCount - 1;

		switch (node->type)
		{
			case kNodeCamera :	free ((NPcameraPtr) node->data); break;
			case kNodeVideo :	free ((NPvideoPtr) node->data); break;
			case kNodeSurface : free ((NPsurfacePtr) node->data); break;
			case kNodePoints :	free ((NPpointsPtr) node->data); break;
			case kNodePin :		free ((NPpinPtr) node->data); break;
			case kNodeGrid :	free ((NPgridPtr) node->data); break;
			default : 
				break;
		}
		data->map.selectedNode = data->map.node[data->map.nodeIndex];
		data->map.activeNode = data->map.selectedNode;
	}

	// if last node deleted then select the camera, note perhaps re-order grid and camera, debug zz
	if ( data->map.nodeIndex < 3)
	{
		data->map.nodeIndex = 1;
		data->map.selectedNode = data->map.node[data->map.nodeIndex];
		data->map.activeNode = data->map.selectedNode;
	}

	free (node);
}




//------------------------------------------------------------------------------
void npInitNodeDefault (NPnodePtr node)
{
	int i = 0;

//	node->id = npNewNodeID();	// called by npNew and preserved during reset

	node->type			= kNodeDefault;		//node type, point, pin, cam, video

	node->data			= NULL;				//node type specific data

	node->selected		= 0;				//defualt is not selected

	node->parent		= NULL;				//parent node, binary tree of nodes
	node->branchLevel	= 0;			//0 is the trunk pin (ice-cream cone)

	for (i=0; i < kNPnodeChildMax; i++)
		node->child[i] = NULL;

	node->childIndex	= 0;
	node->childCount	= 0;				//no children by default

	node->channel.x		= kChannelNull;		//channel to use for x,y,z points
	node->channel.y		= kChannelNull;
	node->channel.z		= kChannelNull;
	
	node->channelIndex.x = 0;				//last updated channel index
	node->channelIndex.y = 0;
	node->channelIndex.z = 0;

	node->averageType.x	= kAverageNull;		//averaging type applied to data
	node->averageType.y	= kAverageNull;
	node->averageType.z	= kAverageNull;
	
	node->sampleInterval.x = 1;				//the interval to be averaged
	node->sampleInterval.y = 1;
	node->sampleInterval.z = 1;

	node->rotate.angle	= 0.0f;				//orientation vector
	node->rotate.x		= 0.0f;
	node->rotate.y		= 0.0f;
	node->rotate.z		= 1.0f;
	
	node->scale.x		= 1.0f;				//scale node XYZ
	node->scale.y		= 1.0f;
	node->scale.z		= 1.0f;
	
	node->translate.x	= 0.0f;				//location relative to nodeOrigin
	node->translate.y	= 0.0f;
	node->translate.z	= 0.0f;

	node->nodeOrigin.x	= 0.0f;				//origin relative to parent origin
	node->nodeOrigin.y	= 0.0f;
	node->nodeOrigin.z	= 0.0f;
	
	node->rotateRate.x	= 0.0f;
	node->rotateRate.y	= 0.0f;
	node->rotateRate.z	= 0.0f;

	node->rotateRad.x	= 0.0f;
	node->rotateRad.y	= 0.0f;
	node->rotateRad.z	= 0.0f;
	
	node->scaleRate.x	= 0.0f;				//scale rate 
	node->scaleRate.y	= 0.0f;
	node->scaleRate.z	= 0.0f;

	node->translateRate.x = 0.0f;			//pans the location of the node, pronunciar en Espanol, lol
	node->translateRate.y = 0.0f;
	node->translateRate.z = 0.0f;

	node->translateVect.x = 0.0f;			//pans the location of the node, pronunciar en Espanol, lol
	node->translateVect.y = 0.0f;
	node->translateVect.z = 0.0f;
	
	node->surfaceType	= 1;//kSurfaceNull;		//FFT layers, mesh, sphere...
//	node->primitive		= 1;														//zz debug
	node->shaderType	= 1;		//wire, flat, gouraud, phong

	node->lineWidth		= 2.0f;				//draws lines, zero for off
	node->pointSize		= 0.0f;				//draws dots, zero for off
	node->circle		= 0.0f;				//draw circle at current data point

	node->colorIndex	= 0;
	
	node->color.r		= 50;				//default color assigned to new data
	node->color.g		= 101;
	node->color.b		= 101;
	node->color.a		= 255;				//fully opaque

	node->colorFade		= false;

	node->hide			= false;			//hide the node, data stays active
	node->freeze		= false;			//freezes data and physics updates
	node->center		= false;			//centers to nodeOrigin
	node->centerPerGrid	= false;			//centers in gridSpace increments

	node->autoZoom.x	= false;			//scales node to fit screen
	node->autoZoom.y	= false;
	node->autoZoom.z	= false;

	node->scroll		= false;			//scroll right to left
	
	node->triggerOnHi.x	= false;			//triggers are multipurpose
	node->triggerOnHi.y	= false;
	node->triggerOnHi.z	= false;

	node->triggerOnLow.x = false;
	node->triggerOnLow.y = false;
	node->triggerOnLow.z = false;
	
	node->triggerLevelHi.x = 0.0f;
	node->triggerLevelHi.y = 0.0f;
	node->triggerLevelHi.z = 0.0f;

	node->triggerLevelLow.x = 0.0f;
	node->triggerLevelLow.y = 0.0f;
	node->triggerLevelLow.z = 0.0f;

	node->triggerCommand = kCtrlCommandNull;	//execute ctrlCommand upon trigger
	
	node->proximity.x	= false;			//used for collision engine 
	node->proximity.y	= false;
	node->proximity.z	= false;
	
	node->proximityRadius.x = 0.0f;
	node->proximityRadius.y = 0.0f;
	node->proximityRadius.z = 0.0f;

	node->drawCurrentValue.x = false;		//draw current values as text
	node->drawCurrentValue.y = false;
	node->drawCurrentValue.z = false;

	node->selectionSet.x	= false;		//default to no selection set
	node->selectionSet.y	= false;
	node->selectionSet.z	= false;

	node->size = sizeof(NPnode);
}


// cameras can be GL scene cameras or real-world cameras
// video is attached as a daughter of the recNode
//-----------------------------------------------------------------------------
void npInitNodeCamera (NPnodePtr node)
{
	NPcameraPtr data = (NPcameraPtr) malloc (sizeof(NPcamera));
	if (data == NULL)
	{
		printf ("\n 4215 error malloc failed cannot write file \n");
		return;
	}

	node->data  = (void*) data;

	node->type			= kNodeCamera;

	data->id			= node->id;
	data->type			= kNodeCamera;

	data->textureID		= 0;					//texture ID in GL


	data->format		= kVideoFormat720p;
	data->interlaced	= kVideoFieldNull;		//progressive scan, no fields
	data->stereo		= kVideoStereoNull;		//not stereoscopic
	data->aspectRatio	= 1.77778f;				//1.77778 is approx 16:9
	data->colorSpace	= kColorSpaceRGB_8;		//8bit RGB is what GPU's output
	data->fps			= 59.97f;
	data->width			= 1280;
	data->height		= 720;

	data->fov			= 35.0f;				//35mm, 70mm...
	data->clipNear		= 1.0f;					//clipping planes
	data->clipFar		= 1000.0f;

	data->aperture		= 1.0f;					//F stop
	data->exposure		= 0.01f;				//in seconds
	data->sensorType	= kCameraSensorNull;	//3CCD, Debayer pattern...

	data->size = sizeof(NPcamera);
}


//-----------------------------------------------------------------------------
void InitNodeVideo (NPnodePtr node)
{
	NPvideoPtr data = (NPvideoPtr) malloc (sizeof(NPvideo));
	if (data == NULL)
	{
		printf ("\n 4214 error malloc failed cannot write file \n");
		return;
	}

	node->data  = (void*) data;

	node->type			= kNodeVideo;

	data->id			= node->id;
	data->type			= kNodeVideo;

	data->textureID		= 0;

	data->format		= kVideoFormat720p;
	data->interlaced	= kVideoFieldNull;		//progressive scan, no fields
	data->stereo		= kVideoStereoNull;		//not stereoscopic
	data->aspectRatio	= 1.77778f;				//1.77778 is approx 16:9
	data->colorSpace	= kColorSpaceRGB_8;		//8bit RGB is what GPU's output
	data->fps			= 59.97f;
	data->width			= 1280;
	data->height		= 720;

	data->bias.r		= 0.0f;				//brightness shift, 0.0 no change
	data->bias.g		= 0.0f;
	data->bias.b		= 0.0f;
	data->bias.a		= 0.0f;
	data->gain.r		= 1.0f;				//color multiplier, 1.0 no change
	data->gain.g		= 1.0f;
	data->gain.b		= 1.0f;
	data->gain.a		= 1.0f;

	data->size = sizeof(NPvideo);
}


//------------------------------------------------------------------------------
void InitNodeSurface (NPnodePtr node)
{
	NPsurfacePtr data = (NPsurfacePtr) malloc (sizeof(NPsurface));
	if (data == NULL)
	{
		printf ("\n 4215 error malloc failed cannot write file \n");
		return;
	}

	node->data  = (void*) data;

	node->type			= kNodeSurface;

	data->id			= node->id;
	data->type			= kNodeSurface;

	data->textureID		= 0;

	data->vertexArray	= NULL;
	data->normalArray	= NULL;
	data->colorArray	= NULL;
	data->vertexCount	= 0;
	data->normalCount	= 0;
	data->arrayIndex	= 0;

	data->size = sizeof(NPsurface);
}


//-----------------------------------------------------------------------------
void InitNodePoints (NPnodePtr node)
{
	NPpointsPtr data = (NPpointsPtr) malloc (sizeof(NPpoints));
	if (data == NULL)
	{
		printf ("\n 4216 error malloc failed cannot write file \n");
		return;
	}

	node->data  = (void*) data;
	
	node->type			= kNodePoints;

	data->id			= node->id;
	data->type			= kNodePoints;

	data->vertexArray	= NULL;				//the array of data points
	data->colorArray	= NULL;				//each point gets color assigned
	data->vertexCount	= 0;				//number of points in the node
	data->arrayIndex	= 0;				//current vertex index in the array

	data->size = sizeof(NPpoints);
}


//-----------------------------------------------------------------------------
void InitNodePin (NPnodePtr node)
{
	NPpinPtr data = (NPpinPtr) malloc (sizeof(NPpin));
	if (data == NULL)
	{
		printf ("\n 4217 error malloc failed cannot write file \n");
		return;
	}

	node->data  = (void*) data;


	node->type			= kNodePin;
	node->surfaceType	= kSurfacePin;		//ice-cream cone shape
	node->shaderType		= kNPprimitivePin;							//zz debug, zz

	data->id			= node->id;
	data->type			= kNodePin;

	data->textureID		= 0;				//0 for no texture assigned

	data->innerRadius	= 0.15f;			//inner radius of the toroid
	data->outerRadius	= 1.5f;			//outer radius of the toroid

	// 4 * outerRadius / innerRadius -> int
	data->radiusRatioIndex = 40;			//sorts pre-loaded GPU toriods
	data->screenSizeIndex = 8;				//same as above for detail level

	data->slices		= 7;				//calculated from screenSizeIndex
	data->stacks		= 16;

	data->translateTex.x = 0.0f;			//position offset of texture map
	data->translateTex.y = 0.0f;
	data->translateTex.z = 0.0f;

	data->scaleTex.x	= 1.0f;
	data->scaleTex.y	= 1.0f;
	data->scaleTex.z	= 1.0f;

	data->rotateTex.x	= 0.0f;
	data->rotateTex.y	= 0.0f;
	data->rotateTex.z	= 0.0f;
	data->rotateTex.angle = 0.0f;			//orientation of texture map

	data->size = sizeof(NPpin);
}


//------------------------------------------------------------------------------
void InitNodeGrid (NPnodePtr node)
{
	NPgridPtr data = (NPgridPtr) malloc (sizeof(NPgrid));
	if (data == NULL)
	{
		printf ("\n 4218 error malloc failed cannot write file \n");
		return;
	}

	node->data  = (void*) data;

	node->type			= kNodeGrid;

	node->colorIndex	= 3;
	node->color.r		= 0;
	node->color.g		= 0;
	node->color.b		= 255;
	node->color.a		= 255;

	data->id			= node->id;
	data->type			= kNodeGrid;

	data->segments.x	= 16;
	data->segments.y	= 16;
	data->segments.z	= 1;
	data->length.x		= 10.0f;
	data->length.y		= 10.0f;
	data->length.z		= 10.0f;

	data->overlay		= false;

	data->size			= sizeof(NPgrid);
}

